/* -*- c -*- 
 * File: pos-sort.c
 * Author: Igor Vlasenko <vlasenko@imath.kiev.ua>
 * Created: Sat Sep 11 19:47:00 2010
 * License: LGPL2+

 * for a given set of numbers x_1 ... x_n and a given set of relations a>b 
 * it computes transitive closure of the relation 
 * and sorts the set of numbers according to the partial order.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>

#include "pstring.h"
#include "loadfile.inc"

// 2^3 = 8 bit in byte
#define LOG2BYTE 3
typedef unsigned int bitpack;
typedef unsigned int dimension;
#define BIT_IN_PACK (8*sizeof(bitpack))
// 2^LOGBITPACK = sizeof(bitpack) bytes
#define LOG2BITPACK (sizeof(bitpack)==1 ? 0 : sizeof(bitpack)==2 ? 1 : sizeof(bitpack)==4 ? 2 : sizeof(bitpack)==8 ? 3 : sizeof(bitpack)==16 ? 4 : 2)

#define PACKSIZE(X) ((X+BIT_IN_PACK-1)>>(LOG2BITPACK+LOG2BYTE))
#define BITHEAD(X) (X>>(LOG2BITPACK+LOG2BYTE))
#define BITTAIL(X) (1<<(X&(BIT_IN_PACK-1)))

inline
static
void* pos_matrix_alloc(const dimension N)
{
  void* matrix = calloc(N,PACKSIZE(N)*sizeof(bitpack));
  if (matrix==NULL) {
    fprintf(stderr,"not enough memory to allocate %ld bytes for dim=%u\n",((long)N)*PACKSIZE(N)*sizeof(bitpack),N);
    exit(EXIT_FAILURE);
  }
#ifdef DEBUG
  fprintf(stderr,"allocated %ld bytes for dim=%u\n",((long)N)*PACKSIZE(N)*sizeof(bitpack),N);
#endif
  return matrix;
}

inline
static
void* pos_matrix_dup(const bitpack* const oldmatrix, const dimension N)
{
  void* matrix = malloc(N*PACKSIZE(N)*sizeof(bitpack));
  if (matrix==NULL) {
    fprintf(stderr,"matrix_dup: not enough memory to allocate %ld bytes for dim=%u\n",((long)N)*PACKSIZE(N)*sizeof(bitpack),N);
    exit(EXIT_FAILURE);
  }
  return memcpy(matrix, oldmatrix, N*PACKSIZE(N)*sizeof(bitpack));
}

inline
static
void* pos_bitvector_dup(const bitpack* const oldvector, const dimension N)
{
  void* vector = malloc(PACKSIZE(N)*sizeof(bitpack));
  if (vector==NULL) {
    fprintf(stderr,"vector_dup: not enough memory to allocate %ld bytes for dim=%u\n",((long) PACKSIZE(N))*sizeof(bitpack),N);
    exit(EXIT_FAILURE);
  }
  return memcpy(vector, oldvector, PACKSIZE(N)*sizeof(bitpack));
}

inline 
static
void pos_matrix_set (bitpack* const matrix, const dimension N, const dimension i, const dimension j)
{
  *(matrix+i*PACKSIZE(N)+BITHEAD(j)) |= BITTAIL(j);
}
inline 
static
void pos_matrix_unset (bitpack* const matrix, const dimension N, const dimension i, const dimension j)
{
  *(matrix+i*PACKSIZE(N)+BITHEAD(j)) &= ~BITTAIL(j);
}
inline
static
bitpack pos_matrix_isset (const bitpack* const matrix, const dimension N, const dimension i, const dimension j)
{
  return *(matrix+i*PACKSIZE(N)+BITHEAD(j)) & BITTAIL(j);
}

inline
static
dimension pos_matrix_pow (const bitpack* const matrix, const dimension N, const dimension i)
{
  dimension j, sum=0;
  for (j=0;j<N;j++)
    {
      if (pos_matrix_isset(matrix,N,i,j)) sum++;
    }
  return sum;
}

struct pos_pow_entry {
  dimension n;
  dimension pow;
};

static
int pos_compar_pow (const void* e1, const void* e2) {
    /* uncomment if you need to list packages in reversed order.
     * return (int) ((const struct pos_pow_entry*)e1)->pow - ((const struct pos_pow_entry*)e2)->pow;
     */
  return (int) ((const struct pos_pow_entry*)e2)->pow - ((const struct pos_pow_entry*)e1)->pow;
}


static
void pos_calculate_transitive_relation(bitpack* const matrix, const dimension N)
{
  int ischanged=1;
  #ifdef DEBUG
  dimension iter=0;
  #endif
  const dimension packsize=PACKSIZE(N);
  // for 0..N
  while (ischanged) {
    register dimension i,j;
    #ifdef DEBUG
    iter++;
    fprintf(stderr, "calculate_transitive_relation: begin iteration %d N=%d packsize=%d\n",iter,N,packsize);
    #endif
    ischanged=0;
    for (i=0;i<N;i++) {
      for (j=0;j<N;j++) {
	// do and (i,j)
	if (pos_matrix_isset(matrix,N,i,j)) {
	  register dimension k;
	  for (k=0;k<packsize;k++) {
	    const register bitpack p=*(matrix+i*packsize+k);
	    const register bitpack q=p | *(matrix+j*packsize+k);
	    *(matrix+i*packsize+k)=q;
	    if (p!=q) ischanged=1;
	  }
	}
      }
    }
  }
  #ifdef DEBUG
  fprintf(stderr, "calculate_transitive_relation took %d iterations\n",iter);
  #endif
}

static
void pos_matrix_art(const bitpack* const matrix, const dimension N)
{
  dimension i,j;
  for (i=0;i<N;i++) putchar('-');
  putchar('\n');
  for (i=0;i<N;i++)
    {
      for (j=0;j<N;j++)
	{
	  if (pos_matrix_isset(matrix,N,i,j)) putchar('1'); else putchar('0');
	}
      putchar('\n');
      }
  for (i=0;i<N;i++) putchar('-');
  putchar('\n');
}

static
void pos_test(void)
{
  dimension N=4;
  bitpack* m = pos_matrix_alloc(N);
  pos_matrix_set(m,N,0,1);
  pos_matrix_set(m,N,1,2);
  pos_matrix_set(m,N,1,3);
  pos_matrix_art(m,N);
  pos_calculate_transitive_relation(m,N);
  pos_matrix_art(m,N);
  dimension j;
  for (j=0;j<N;j++)
    {
      printf("%d pow=%d\n", j, pos_matrix_pow(m,N,j));
    }
}

inline
static
void* pos_bitvector_alloc(const dimension N)
{
  return calloc(PACKSIZE(N),sizeof(bitpack));
}

inline
static
void* pos_bitvector_erase(bitpack* const vector, const dimension N)
{
  return memset(vector,0,PACKSIZE(N)*sizeof(bitpack));
}

inline 
static
void pos_bitvector_set (bitpack* const vector, const dimension i)
{
  *(vector+BITHEAD(i)) |= BITTAIL(i);
}
inline
static
bitpack pos_bitvector_isset (const bitpack* const vector,const dimension i)
{
  return *(vector+BITHEAD(i)) & BITTAIL(i);
}

static
bitpack* pos_matrix_slice(const bitpack* const oldmatrix, const dimension N, const bitpack* const slicevector, const dimension NR)
{
  void* matrix = pos_matrix_alloc(NR);
  dimension i,j;
  dimension newi=0,newj;
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(slicevector,i)) {
      newj=0;
      for (j=0;j<N;j++) {
	if (pos_bitvector_isset(slicevector,j)) {
	  if (pos_matrix_isset(oldmatrix,N,i,j)) {
	    pos_matrix_set(matrix,NR,newi,newj);
	  }
	  newj++;
	}
      }
      newi++;
    }
  }
  return matrix;
}

static
bitpack* pos_bitvector_slice(const bitpack* const oldvector, const dimension N, const bitpack* const slicevector, const dimension NR)
{
  void* bitvector = pos_bitvector_alloc(NR);
  register dimension i;
  dimension newi=0;
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(slicevector,i)) {
      if (pos_bitvector_isset(oldvector,i)) {
	pos_bitvector_set(bitvector,newi);
      }
      newi++;
    }
  }
  return bitvector;
}

static
void pos_bitvector_fill_from (FILE* in, bitpack* const vector, const dimension N) {
  dimension i;
  while (fscanf(in,"%u",&i)==1) {
    if (i>=N) {
      fprintf(stderr,"incorrect input: %d is greater than %d\n",i, N-1);
      exit(2);
    }
    //fprintf(stderr,"bitvector scanf:%u\n",i);
    pos_bitvector_set(vector,i);
  }
}

static
void pos_matrix_fill_from (FILE* in, bitpack* const m, const dimension N) {
  dimension i,j;
  while (fscanf(in,"%u %u",&i,&j)==2) {
    if (i>=N || j>=N) {
      fprintf(stderr,"incorrect input: %d %d are greater than %d\n",i, j, N-1);
      exit(2);
    }
    //fprintf(stderr,"matrix scanf:%u %u\n",i,j);
    pos_matrix_set(m,N,i,j);
  }
}

static
bitpack* pos_bitvector_from_file (const char* filename, const dimension N) {
  bitpack* content = pos_bitvector_alloc(N);
  if (filename!=NULL) {
    FILE* IN = fopen(filename, "r");
    if (NULL==IN) {
      perror("can't open file :");
      fprintf(stderr,"opening %s failed.\n",filename);
      exit(EXIT_FAILURE);
    }
    pos_bitvector_fill_from(IN,content,N);
    fclose(IN);
  }
  return content;
}

inline 
static
void pos_bitvector_complement (bitpack* const vector, const dimension N)
{
  dimension i;
  for (i=0;i<PACKSIZE(N);i++) {
    *(vector+i)=~*(vector+i);
  }
}

inline 
static
dimension pos_bitvector_sum (bitpack* const vector, const dimension N)
{
  dimension i;
  dimension sum = 0;
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(vector,i)) sum++;
  }
  return sum;
}

static
void
pos_load_string_map (PSTRING memarea, PSTRING* map, const char* filename, const dimension N) {
  PSTRING str={memarea.begin,memarea.begin};
  dimension counter=0;
  const char* position=memarea.begin;
  while (position<memarea.endnext && counter<N) {
    if ('\n'==*position) {
      str.endnext=position;
      *map=str;
      map++;
      counter++;
      str.begin=str.endnext=position+1;
    }
    position++;
  }
  if ((memarea.begin <= position-1) && ('\n'==*(position-1))) position--;
  if (str.begin<position) {
      str.endnext=position;
      *map=str;
      map++;
      counter++;
  }
  if (counter<N) {
    fprintf(stderr, "invalid map file %s: got %d lines, expected at least %d lines.\n", filename,counter, N);
    exit(EXIT_FAILURE);
  }
}

static
void
set_cycle_component_and_current_cycle (bitpack* current_cycle, bitpack* cycle_component_mark, const bitpack* const m, const dimension N, const dimension i) {
  dimension j;
  pos_bitvector_erase(current_cycle, N);
  for (j=0;j<N;j++) {
    if (pos_matrix_isset(m,N,i,j) &&
	pos_matrix_isset(m,N,j,i)) {
      pos_bitvector_set(cycle_component_mark,j);
      pos_bitvector_set(current_cycle,j);
    }
  }
}

static
int 
pos_is_marked_srpm_in_cycle (bitpack* current_cycle, const bitpack* const marked_srpm_flag, const dimension N) {
  dimension j;
  for (j=0;j<N;j++) {
    if (pos_bitvector_isset(current_cycle,j) && pos_bitvector_isset(marked_srpm_flag,j)) return 1;
  }
  return 0;
}

static
bitpack* 
pos_list_imporant_rpms (const bitpack* const an_rpm_flag, const bitpack* const marked_rpm_flag, const bitpack* const matrix, const dimension N) {
  dimension i,j;
  bitpack* rpms_matrix = pos_matrix_alloc(N);
  bitpack* important_rpm = pos_bitvector_alloc(N);
  /* fill up cycle rpm matrix (w/o srpms) */
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(an_rpm_flag,i)) {
      for (j=0;j<N;j++) {
	if (pos_matrix_isset(matrix,N,i,j) &&
	    pos_bitvector_isset(an_rpm_flag,j)
	    ) {
    	  pos_matrix_set(rpms_matrix,N,i,j);
	}
      }
    }
  }
  pos_calculate_transitive_relation(rpms_matrix,N);
  /* mark up rpms that will require marked rpms */
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(marked_rpm_flag,i)) {
      pos_bitvector_set(important_rpm,i);
      for (j=0;j<N;j++) {
	if (pos_matrix_isset(rpms_matrix,N,j,i)
	    ) {
	  pos_bitvector_set(important_rpm,j);
#ifdef DEBUG2
	  fprintf(stderr,"I: %d is important due to %d\n",j,i);
#endif
	}
      }
    }
  }
  free(rpms_matrix);
  return important_rpm;
}

static
void pos_bitvector_debug(const bitpack* vector, const dimension N, const dimension* const node_map) {
  dimension idebug;
  for (idebug=0;idebug<N;idebug++) {
    if (pos_bitvector_isset(vector,idebug)) {
      fprintf(stderr," %d",idebug);
    }
  }
  fprintf(stderr,"\n");
  fprintf(stderr,"unmapped values: ");
  for (idebug=0;idebug<N;idebug++) {
    if (pos_bitvector_isset(vector,idebug)) {
      fprintf(stderr," %d",*(node_map+idebug));
    }
  }
  fprintf(stderr,"\n");
}

static
inline
void pos_output_node(FILE* OUT, const dimension node, PSTRING* const strings_map
) {
  if (strings_map) {
    PSTRING str = *(strings_map+node);
    fprintf(OUT,"%.*s\n",(int) (str.endnext-str.begin), str.begin);
  } else {
    fprintf(OUT,"%d\n", node);
  }
}

static
inline
void pos_output_bitvector(FILE* OUT, const bitpack* const vector, 
			  const dimension N, PSTRING* const strings_map
) {
  dimension i;
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(vector,i)) {
      pos_output_node(OUT, i, strings_map);
    }
  }
}

static
inline
void pos_output_mapped(FILE* OUT, const bitpack* const restriction_vector, 
		       const dimension node, 
		       const dimension* const node_map,
		       PSTRING* const strings_map
) {
  if (restriction_vector==NULL || pos_bitvector_isset(restriction_vector,node)) {
    dimension real_node=*(node_map+node);
    pos_output_node(OUT, real_node, strings_map);
  }
}

static
inline
void pos_output_bitvector_mapped(FILE* OUT, const bitpack* const restriction_vector, 
			  const bitpack* current_cycle, 
			  const dimension N,
			  const dimension* const node_map,
			  PSTRING* const strings_map
) {
  dimension i;
  for (i=0;i<N;i++) {
    if (pos_bitvector_isset(current_cycle,i)) {
      pos_output_mapped(OUT, restriction_vector,i, node_map, strings_map);
    }
  }
}

static
void pos_print_cycle_component_file(
		     const dimension N,
		     const dimension* const node_map,
		     PSTRING* cycle_strings_map,
		     const char* cycle_component_template,
		     dimension* component_counter_ref,
		     bitpack* current_cycle
) {
	if (cycle_component_template) {
		char cycle_component_file [strlen(cycle_component_template)+40];
		FILE* CYCLE_OUT;
		sprintf(cycle_component_file, "%s%05d", cycle_component_template,*component_counter_ref);
		CYCLE_OUT = fopen(cycle_component_file, "w");
		if (NULL==CYCLE_OUT) {
			perror("can't open cycle info file:");
			exit(EXIT_FAILURE);
		}
		pos_output_bitvector_mapped(CYCLE_OUT, NULL, current_cycle, N, node_map, cycle_strings_map);
		fclose(CYCLE_OUT);
	}
}

static
void pos_sort_matrix(const bitpack* const m0, const dimension N,
		     const dimension previous_N,
		     const bitpack* const an_rpm_flag, 
		     const bitpack* const marked_rpm_flag, 
		     const bitpack* const marked_srpm_flag,
		     const dimension* const node_map,
		     PSTRING* output_strings_map,
		     PSTRING* cycle_strings_map,
		     int do_breaking_cycles_through_unmarked,
		     FILE* OUT,
		     const char* cycle_component_template,
		     dimension* component_counter_ref
) {
  bitpack* const m= pos_matrix_dup(m0, N);

  /* calculate transitive relation */

#ifdef DEBUG_MATRIX_ART
  pos_matrix_art(m,N);
#endif
  pos_calculate_transitive_relation(m,N);
#ifdef DEBUG_MATRIX_ART
  pos_matrix_art(m,N);
#endif

  /* sort */

  struct pos_pow_entry* pos_pow_array = calloc (N,sizeof(struct pos_pow_entry));
  dimension k;
  for (k=0;k<N;k++) {
    dimension pw = pos_matrix_pow(m,N,k);
    struct pos_pow_entry pw_entry = {k, pw};
    pos_pow_array[k]=pw_entry;
    //printf("%d pow=%d\n", k, pw);
  }
  qsort(pos_pow_array,N,sizeof(struct pos_pow_entry),&pos_compar_pow);

  /* and deal with possible cycles ... */

  bitpack* cycle_component_mark = pos_bitvector_alloc(N);
  bitpack* current_cycle = pos_bitvector_alloc(N);

  for (k=0;k<N;k++) {
    dimension node = pos_pow_array[k].n;
    /* debug */
    //PSTRING str = *(output_strings_map+node);
    //fprintf(stderr, "%d (%.*s) pow=%d\n", pos_pow_array[k].n,(int) (str.endnext-str.begin), str.begin, pos_pow_array[k].pow);
    /* debug if output_strings_map is not defined */
    //fprintf(stderr, "%d pow=%d\n", pos_pow_array[k].n, pos_pow_array[k].pow);
    
    /* already marked (processed) connected component of cycle set */
    if (pos_bitvector_isset(cycle_component_mark,node)) continue;

    /* normal node, does not belong to a cycle set */
    if (!pos_matrix_isset(m,N,node,node)) {
      pos_output_mapped(OUT, marked_srpm_flag, node, node_map, output_strings_map);
      continue;
    }
      
    /* node belongs to a cycle set */
    //fprintf(stderr, "k=%d entering cycle\n", k);
    set_cycle_component_and_current_cycle(current_cycle,cycle_component_mark,m,N,node);

    if (marked_srpm_flag==NULL) {
      //fprintf(stderr, "k=%d plain cycle -> output\n", k);
      /* all elements of (unmarked) cycle have the same pow */
      pos_output_bitvector_mapped(OUT, marked_srpm_flag, current_cycle, N, node_map, output_strings_map);
      continue;
    } else if (!pos_is_marked_srpm_in_cycle(current_cycle,marked_srpm_flag,N)) {
      /* the cycle has no marked srpm - do nothing */
      continue;
    } else {
      /* marked cycle found */
      dimension NR = pos_bitvector_sum(current_cycle,N);
      //fprintf(stderr, "k=%d entered cycle else\n", k);
	
      if (
	  /* recurrent calls does not break the loop; giving up */
	  (NR==N && N==previous_N) /* cycle is unbreakable */
	  || !do_breaking_cycles_through_unmarked
	      ) {
	  //fprintf(stderr, "k=%d inside NR==N\n", k);
#ifdef DEBUG2
	  fprintf(stderr,"reporting cycle %d of len=%d detected at %d:\n",*component_counter_ref,NR,node);
	  pos_bitvector_debug(current_cycle,N,node_map);
#endif
	  (*component_counter_ref)++;
	  pos_output_bitvector_mapped(OUT, marked_srpm_flag, current_cycle, N, node_map, output_strings_map);
	  pos_print_cycle_component_file(N, node_map, cycle_strings_map, cycle_component_template, component_counter_ref, current_cycle);
	  continue;
      } else {
	  /* we try to break the cycle, dropping relatins to unmarked nodes. */

	  //fprintf(stderr, "k=%d do_breaking_cycles_through_unmarked (else)\n", k);
		  
#ifdef DEBUG
	  fprintf(stderr,"cycle detected with dim=%d. Trying to break the cycle...\n", NR);
#endif
	  bitpack* mnew= pos_matrix_slice(m0, N, current_cycle, NR);
	  bitpack* sliced_an_rpm_flag = pos_bitvector_slice(an_rpm_flag,N, current_cycle, NR);
	  bitpack* sliced_marked_rpm_flag = pos_bitvector_slice(marked_rpm_flag,N, current_cycle, NR);
	  bitpack* sliced_marked_srpm_flag = pos_bitvector_slice(marked_srpm_flag,N, current_cycle, NR);
		      
	  dimension* sliced_node_map = malloc(NR*sizeof(dimension));
	  dimension i, ii=0;
	  for (i=0;i<N;i++) {
	      if (pos_bitvector_isset(current_cycle,i)) {
		  *(sliced_node_map+(ii++))=*(node_map+i);
	      }
	  }
	  bitpack* sliced_important_rpm = pos_list_imporant_rpms(sliced_an_rpm_flag,sliced_marked_rpm_flag,mnew,NR);

#ifdef DEBUG2
	  fprintf(stderr,"important rpm: ");
	  pos_bitvector_debug(sliced_important_rpm,NR, sliced_node_map);
#endif

	  dimension j;
	  for (j=0;j<NR;j++) {
	      if ( /* j is srpm */
		  !pos_bitvector_isset(sliced_an_rpm_flag,j)) {
		  if (!pos_bitvector_isset(sliced_marked_srpm_flag,j)) {
		      /* break cycles on non-important srpms */
		      register dimension k;
		      for (k=0;k<NR;k++) {
			  pos_matrix_unset(mnew,NR,j,k);
			  pos_matrix_unset(mnew,NR,k,j);
		      }
		  } else {
		      /* srpm is important; break cycles on non-important rpms */
		      
		      register dimension k;
		      for (k=0;k<NR;k++) {
			  /* k is rpm */
			  /* not necessary -- srpm does not depend on other srpms 
			     if (pos_bitvector_isset(sliced_an_rpm_flag,j) */
			  /* no need to check --  it is faster to just unset regardless.
			     && pos_matrix_isset(mnew,NR,j,k) or k,j 
			     ) {*/
			  if (!pos_bitvector_isset(sliced_important_rpm,k)) {
			      /* remove dependency as it is neglectable, 
				 effectively breaking the cycle */
			      pos_matrix_unset(mnew,NR,j,k);
			      /* also remove child rpm from dependencies -- it is not important */
			      pos_matrix_unset(mnew,NR,k,j);
			      /*
				fprintf(stderr,"breaking the circle: unsetting dep: %d > %d\n",j,k);
				} else {
				fprintf(stderr,"can't break the circle: %d > %d - important rpm flag\n",j,k);
			      */
			  }
		      }
		  }
	      }
	  }
	  free(sliced_important_rpm);
		  
	  pos_sort_matrix(mnew, NR, N,
			  sliced_an_rpm_flag, 
			  sliced_marked_rpm_flag, 
			  sliced_marked_srpm_flag,
			  sliced_node_map,
			  output_strings_map,
			  cycle_strings_map,
			  do_breaking_cycles_through_unmarked,
			  OUT,
			  cycle_component_template,
			  component_counter_ref);
	  
	  free(sliced_node_map);
	  free(mnew);
	  free(sliced_an_rpm_flag);
	  free(sliced_marked_rpm_flag);
	  free(sliced_marked_srpm_flag);
      }
    }
  } /* end for */

  free(cycle_component_mark);
  free(current_cycle);
  return;
}

/* 
 *  unset build-from arrows. set Req: and BReq:. 
 * pos_transitive_greater:
 *  calculate which srpms have the rpms from 
 *  the given set of rpm in their build environment.
 */
static
void pos_transitive_greater(const bitpack* const m0, const dimension N,
		     const bitpack* const an_rpm_flag,
		     const bitpack* const marked_rpm_flag,
		     PSTRING* output_strings_map,
		     FILE* OUT
) {
  bitpack* const m= pos_matrix_dup(m0, N);
  bitpack* const who_greater= pos_bitvector_alloc(N);
  dimension j, i;

  //pos_matrix_art(m,N);
  pos_calculate_transitive_relation(m,N);
  //pos_matrix_art(m,N);

  for (j=0;j<N;j++)
    {
      if (marked_rpm_flag && ! pos_bitvector_isset(marked_rpm_flag,j)) continue;
      for (i=0;i<N;i++)
	{
	    if (!(an_rpm_flag && pos_bitvector_isset(an_rpm_flag,i)) && pos_matrix_isset(m,N,j,i)) pos_bitvector_set(who_greater,i);
	}
    }
  pos_output_bitvector(OUT, who_greater, N, output_strings_map);
}

/*  TODO:
 *  unset build-from arrows. set Req: and BReq:. 
 * pos_transitive_less:
 *  calculate the union of build environments (in intersection with 
 *  the given set of rpms) for the given set of srpms.
 */
static
void pos_transitive_less(const bitpack* const m0, const dimension N,
		     const bitpack* const an_rpm_flag,
		     const bitpack* const marked_rpm_flag,
		     PSTRING* output_strings_map,
		     FILE* OUT
) {
  bitpack* const m= pos_matrix_dup(m0, N);
  bitpack* const who_greater= pos_bitvector_alloc(N);
  // dimension j, i;

  //pos_matrix_art(m,N);
  pos_calculate_transitive_relation(m,N);
  //pos_matrix_art(m,N);

  /* TODO:
  for (j=0;j<N;j++)
    {
      if (marked_rpm_flag && ! pos_bitvector_isset(marked_rpm_flag,j)) continue;
      for (i=0;i<N;i++)
	{
	  if (!(an_rpm_flag && pos_bitvector_isset(an_rpm_flag,i)) && pos_matrix_isset(m,N,i,j)) pos_bitvector_set(who_greater,i);
	}
    }
  */
  pos_output_bitvector(OUT, who_greater, N, output_strings_map);
}



static
void pos_sort_usage(void)
{
  fprintf(stderr,"usage: %s [-c] [-C cycleprefix] [-r file] [-s file] [-R file] [-i infile] [-o outfile] <num of elements>\n"
	  "options:\n"
	  "\t-c\texit if cycle detected\n"
	  "\t-C <tmpl>\tprefix for cycle information files\n"
	  "\t-s <file>\ttransaction srpms\n"
	  "\t-r <file>\ttransaction rpms\n"
	  "\t-R <file>\tall rpms.  Alternatively, -S can be used.\n"
	  "\t-S <file>\tall srpms. Alternatively, -R can be used.\n"
	  "\t-i <file>\tinput file\n"
	  "\t-o <file>\toutput file\n","girar-nmu-pos-sort");
 }

int main(int argc, char** argv)
{
  dimension N0=0;
  int exit_on_cycles=0;
  int opt_break_cycles_through_unmarked=0;
  const char* cycle_component_template=NULL;
  const char* infile=NULL;
  const char* outfile=NULL;
  const char* marked_srpms_file=NULL;
  const char* marked_rpms_file=NULL;
  const char* all_rpms_file=NULL;
  const char* all_srpms_file=NULL;
  const char* cycle_strings_file=NULL;
  const char* output_strings_file=NULL;
  FILE* OUT = NULL;
  int mode_greater=0;

  //printf("argc=%d\n", argc);
  if (argc<2) {
    fprintf(stderr,"not enough arguments.\n");
    pos_sort_usage();
    exit(EXIT_FAILURE);
  }

  static struct option long_options[] = {
    {"infile", 1, 0, 'i'},
    {"outfile", 1, 0, 'o'},
    {"greater", 0, 0, 'g'},
    {"marked-rpms", 1, 0, 'r'},
    {"marked-srpms", 1, 0, 's'},
    {"break-cycles-through-unmarked", 0, 0, 'b'},
    {"no-break-cycles-through-unmarked", 0, 0, 'B'},
    {"help", 0, 0, '?'},
    {0, 0, 0, 0}
  };

  int opt;
  //  while ((opt = getopt(argc, argv, "bBcC:gi:m:M:o:r:R:s:S:")) != -1) {
  while ((opt = getopt_long(argc, argv, "bBcC:gi:m:M:o:r:R:s:S:", long_options, NULL)) != -1) {
    switch (opt) {
    case 'b':
      opt_break_cycles_through_unmarked=1;
      break;
    case 'B':
      opt_break_cycles_through_unmarked=0;
      break;
    case 'c':
      exit_on_cycles=1;
      break;
    case 'C':
      cycle_component_template=optarg;
      break;
    case 'i':
      infile=optarg;
      break;
    case 'g':
      mode_greater=1;
      break;
    case 'o':
      outfile=optarg;
      break;
    case 'm':
      output_strings_file=optarg;
      break;
    case 'M':
      cycle_strings_file=optarg;
      break;
    case 'r':
      marked_rpms_file=optarg;
      break;
    case 'R':
      all_rpms_file=optarg;
      break;
    case 's':
      marked_srpms_file=optarg;
      break;
    case 'S':
      all_srpms_file=optarg;
      break;
    default: /* '?' */
    pos_sort_usage();
    exit(EXIT_FAILURE);
    }
  }
  
  if (optind >= argc) {
    fprintf(stderr, "Expected <num of elements> argument after options\n");
    exit(EXIT_FAILURE);
  }

  if ((marked_rpms_file !=NULL)&& all_rpms_file==NULL && all_srpms_file==NULL ) {
    fprintf(stderr, "-r option requires either -R or -S option\n");
    exit(EXIT_FAILURE);
  }

  sscanf(argv[optind],"%u",&N0);
  if (N0<=0) {
    fprintf(stderr,"incorrect first non-option argument. Should be a number.\n");
    pos_sort_usage();
    exit(1);
  }
  const dimension N=N0;
  bitpack* marked_rpm_flag = pos_bitvector_from_file(marked_rpms_file, N);
  bitpack* marked_srpm_flag = pos_bitvector_from_file(marked_srpms_file, N);
  bitpack* an_rpm_flag = pos_bitvector_from_file(all_rpms_file, N);

  if (all_rpms_file==NULL && all_srpms_file!=NULL) {
    free(an_rpm_flag);
    an_rpm_flag = pos_bitvector_from_file(all_srpms_file, N);
    pos_bitvector_complement(an_rpm_flag,N);
  }

  bitpack* m0 = pos_matrix_alloc(N);
  if (infile) {
    FILE* IN = fopen(infile, "r");
    if (NULL==IN) {
      perror("can't open matrix infile:");
      exit(EXIT_FAILURE);
    }
    pos_matrix_fill_from(IN,m0,N);
    fclose(IN);
  } else {
    pos_matrix_fill_from(stdin,m0,N);
  }

  if (outfile) {
    OUT = fopen(outfile, "w");
    if (NULL==OUT) {
      perror("can't open output file:");
      exit(EXIT_FAILURE);
    }
  } else OUT=stdout;

  PSTRING output_strings_memarea = {NULL,NULL};
  PSTRING cycle_strings_memarea = {NULL,NULL};
  PSTRING* output_strings_map = NULL;
  PSTRING* cycle_strings_map = NULL;
  if (output_strings_file) {
    output_strings_memarea = mmap_load_file (output_strings_file);
    output_strings_map = calloc(N,sizeof(PSTRING));
    pos_load_string_map (output_strings_memarea, output_strings_map, output_strings_file, N);
  }
  if (cycle_strings_file) {
    cycle_strings_memarea = mmap_load_file (cycle_strings_file);
    cycle_strings_map = calloc(N,sizeof(PSTRING));
    pos_load_string_map (cycle_strings_memarea, cycle_strings_map, cycle_strings_file, N);
  }

  if (mode_greater) {
    pos_transitive_greater(m0, N, an_rpm_flag, marked_rpm_flag, output_strings_map, OUT);
  } else {
    dimension component_counter = 0;

    dimension* node_map = malloc(N*sizeof(dimension));
    dimension i;
    for (i=0;i<N;i++) {
      /* identity map */
      *(node_map+i)=i;
    }

    pos_sort_matrix(m0, N, N+1,
		    an_rpm_flag, 
		    marked_rpm_flag, 
		    marked_srpm_flag,
		    node_map,
		    output_strings_map,
		    cycle_strings_map ? cycle_strings_map : output_strings_map,
		    opt_break_cycles_through_unmarked,
		    OUT,
		    cycle_component_template,
		    &component_counter
		    );
    free(node_map);
    if (exit_on_cycles && component_counter>0){
      fprintf(stderr,"%d bloking cycles found.\n",component_counter);
      if (cycle_component_template) 
	fprintf(stderr,"Cycle content is saved in %sXXXXX files.\n",cycle_component_template);
      exit(3);
    }
  }
  if (outfile) fclose(OUT);

  free(m0);
  if (output_strings_memarea.begin) mmap_unload_file (output_strings_memarea);
  if (cycle_strings_memarea.begin) mmap_unload_file (cycle_strings_memarea);

  return 0;
}




/*
 *  Local Variables:
 *  mode: c
 *  End:
 */
